package Apache::AxKit::Language::FOP;

use strict;
use Apache ();
use Apache::Constants ();
use Apache::AxKit::Language ();
use Apache::AxKit::Cache ();
use Digest::MD5 ();
use XML::LibXML ();
use URI ();
use URI::file ();
use File::Spec ();
use File::Temp ();
use File::Path ();
use File::stat ();
use File::Basename ();
use Fcntl qw( :DEFAULT :mode :flock );
use SDBM_File;
@Apache::AxKit::Language::FOP::ISA = qw( Apache::AxKit::Language );

$Apache::AxKit::Language::FOP::VERSION = '0.92';

my $content_type = {
		    pdf => 'application/pdf',
		    mif => 'application/vnd.mif',
		    pcl => 'application/vnd.hp-PCL',
		    ps => 'application/postscript',
		    txt => 'text/plain',
		    svg => 'image/svg+xml',
		    at => 'text/xml',
		   };

sub stylesheet_exists { 0 }

#-------------------------------------------------------------------------------
sub handler {
    my $class = shift;
    my( $r, $xml_provider, undef, $last_in_chain ) = @_;

    my $dom = get_dom( $r, $xml_provider );
    my( $config, @base ) = get_config( $r );
    my $graphics = resolve_external_graphics( $dom, @base, $xml_provider );
    AxKit::add_depends( "{content}$_" ) for @$graphics;
    my $input = $dom->toString;
    my $style = $AxKit::Cfg->PreferredStyle;
    unless ( defined $content_type->{$style} ) {
	AxKit::Debug( 1, "AxKit FOP: style '$style' unrecognized; using 'pdf'" );
	$style = 'pdf';
    }

    # Check if output is in the cache and does not need to be recreated.
    my( $cache, $recreate, $stats ) =
      caching( $r, \$input, $style, $graphics, $xml_provider );
    my $fh;
    if ( $cache and $cache->exists and not $recreate ) {
	AxKit::Debug( 1, "AxKit FOP: delivering output from cache for " .
		      $xml_provider->key );
	$fh = $cache->get_fh;
    }
    else {
	my $output_file = run_fop( $r, $style, \$input, $config,
				   $graphics, $stats, $xml_provider );
	$fh = Apache->gensym;
	open( $fh, "<$output_file" )
	  or die "AxKit FOP: could not open $output_file: $!";
	if ( $cache ) {
	    # Cache the output for future requests.
	    AxKit::Debug( 1, "AxKit FOP: saving output in cache for " .
			  $xml_provider->key );
	    $cache->write( do { local $/; <$fh> } );
	    seek( $fh, 0, 0 );
	}
    }

    # Read output and send it.
    $AxKit::Cfg->AllowOutputCharset(0);
    $r->content_type( $content_type->{$style} );
    $r->print( do { local $/; <$fh> } );
    close( $fh ) or die "AxKit FOP: could not close output filehandle: $!";

    return Apache::Constants::OK;
}

#-------------------------------------------------------------------------------
# Get already parsed XSL-FO input or find it, parse it, and return DOM.
#-------------------------------------------------------------------------------
sub get_dom {
    my( $r, $xml_provider ) = @_;

    # Check if XSL-FO input has already been parsed.
    if ( my $dom = $r->pnotes( 'dom_tree' ) ) {
	delete $r->pnotes->{'dom_tree'};
	AxKit::Debug( 10, "AxKit FOP: input from dom_tree" );
	return $dom;
    }

    # Get XSL-FO input.
    my( $string, $string_ref, $fh );
    if ( $string = $r->pnotes( 'xml_string' ) ) {
	delete $r->pnotes->{'xml_string'};
	AxKit::Debug( 10, "AxKit FOP: input from xml_string" );
    }
    else {
	eval {
	    $string_ref = $xml_provider->get_strref;
	    AxKit::Debug( 10, "AxKit FOP: input from get_strref" );
	};
	if ( $@ ) {
	    $fh = $xml_provider->get_fh;
	    AxKit::Debug( 10, "AxKit FOP: input from get_fh" );
	}
    }

    # Parse XSL-FO input.
    my $parser = new XML::LibXML;
    my $dom = eval {
	if ( $fh ) {
	    $parser->parse_fh( $fh );
	}
	else {
	    $parser->parse_string( $string || $$string_ref );
	}
    };
    if ( $@ ) {
	die "AxKit FOP: could not parse XSL-FO from " . $xml_provider->key .
	  ": $@";
    }
    $dom;
}

#-------------------------------------------------------------------------------
# Find existing FOP configuration file or create a new one with
# baseDir set to source directory.  Return source directory and
# baseDir (if different from source directory).
#-------------------------------------------------------------------------------
sub get_config {
    my $r = shift;

    my $source_dir = $r->filename;
    $source_dir = File::Basename::dirname( $source_dir ) unless -d $source_dir;
    AxKit::Debug( 8, "AxKit FOP: source directory is $source_dir" );

    # Parse existing FOP configuration file or create a new DOM of one.
    my $config_file = $r->dir_config( 'AxFOPConfigFile' );
    my $dom;
    if ( defined $config_file and -e $config_file ) {
	my $parser = new XML::LibXML;
	$dom = eval { $parser->parse_file( $config_file ) };
	die "AxKit FOP: could not parse $config_file: $@" if $@;
	# Check for existing baseDir entry in configuration.
	my $xpath = "/configuration/entry[key='baseDir']/value";
	( my $base_dir ) = $dom->findnodes( $xpath );
	if ( defined $base_dir ) {
	    $base_dir = $base_dir->textContent;
	    AxKit::Debug( 8, "AxKit FOP: baseDir is $base_dir in" .
			  " existing config file $config_file" );
	    return ( $config_file, $source_dir, $base_dir );
	}
    }
    else {
	warn "AxKit FOP: AxFOPConfigFile does not exist: $config_file"
	  if $config_file;
	$dom = XML::LibXML->createDocument;
	$dom->setStandalone(1);
	$dom->setDocumentElement( $dom->createElement( 'configuration' ) );
    }

    # Add an entry for baseDir set to source directory.
    my $entry = $dom->createElement( 'entry' );
    $dom->getDocumentElement->appendChild( $entry );
    $entry->appendTextChild( key => 'baseDir' );
    $entry->appendTextChild( value => $source_dir );
    AxKit::Debug( 8, "AxKit FOP: setting baseDir to $source_dir" );

    return ( $dom, $source_dir, undef );
}

#-------------------------------------------------------------------------------
# Search for external-graphics, resolve them relative to source
# directory or baseDir, and return those defined in file system.
#-------------------------------------------------------------------------------
sub resolve_external_graphics {
    my( $input, $source_dir, $base_dir, $xml_provider ) = @_;

    # Get list of fo:external-graphics defined in document.
    my $root = $input->getDocumentElement;
    unless ( defined $root ) {
	warn "AxKit FOP: could not find root element from " .
	  $xml_provider->key;
	return;
    }
    my $prefix =
      $root->lookupNamespacePrefix( 'http://www.w3.org/1999/XSL/Format' );
    unless ( defined $prefix ) {
	warn "AxKit FOP: could not find prefix for XSL-FO from " .
	  $xml_provider->key;
	return;
    }
    my $xpath = "//$prefix:external-graphic[\@src]/\@src";
    my @srcs = $root->findnodes( $xpath );

    # Resolve src of external-graphics relative first to source_dir
    # and second to base_dir.  Return list of external-graphics
    # located in file system.
    local $URI::ABS_ALLOW_RELATIVE_SCHEME = 1;
    my $source_uri = URI::file->new( $source_dir );
    $source_uri .= '/' if substr( $source_uri, 1 ) ne '/';
    my $base_uri = defined $base_dir ? URI::file->new( $base_dir ) : undef;
    $base_uri .= '/' if defined $base_uri and substr( $base_uri, 1 ) ne '/';
    my %graphics;
    foreach my $src ( @srcs ) {
	my $rel_uri = URI->new( $src->getValue, $source_uri );
	my $abs_uri = $rel_uri->abs( $source_uri ); # resolve relative to source
	if ( $base_uri and $abs_uri->scheme eq 'file'
	     and not -e $abs_uri->file )
	{
	    $abs_uri = $rel_uri->abs( $base_uri ); # resolve relative to baseDir
	}
	unless ( $rel_uri->as_string eq $abs_uri->as_string ) {
	    $src->setValue( $abs_uri->as_string );
	    AxKit::Debug( 4, "AxKit FOP: resolving $rel_uri => $abs_uri" );
	}
	++$graphics{ $abs_uri->file } if $abs_uri->scheme eq 'file';
    }
    [ keys %graphics ];
}

#-------------------------------------------------------------------------------
# Determine whether or not to cache and whether cache is up-to-date.
#-------------------------------------------------------------------------------
sub caching {
    my( $r, $input, $style, $graphics, $xml_provider ) = @_;

    # Determine if we should cache separately or leave AxKit to do it.
    my $axfopcache = lc $r->dir_config( 'AxFOPCache' ) || '';
    if ( $axfopcache eq 'ifnocache' ) {
	$axfopcache = $AxKit::Cache->no_cache ? 1 : 0;
    }
    elsif ( $axfopcache eq 'on' ) {
	$axfopcache = 1;
    }
    else {
	$axfopcache = 0;
    }
    AxKit::Debug( 8, "AxKit FOP: AxFOPCache = $axfopcache" );
    return unless $axfopcache and not $AxKit::Cfg->NoCache;

    # Create a cache object.
    my $digest = Digest::MD5->new;
    $digest->add( $$input );
    $digest->add( $style );
    my $extra = $digest->hexdigest;
    my $cache = Apache::AxKit::Cache->new( $r, $xml_provider->key, $extra );

    # Check modification-timestamps of dependencies to see if any have
    # been updated since the output was cached.
    return ( $cache ) unless $AxKit::Cfg->DependencyChecks; # do not check
    return ( $cache ) unless $cache and $cache->exists; # cache not up-to-date
    return ( $cache ) unless $graphics and @$graphics; # cache up-to-date
    my $stats = {};
    my $mtime = $cache->mtime;
    foreach my $graphic ( @$graphics ) {
	AxKit::Debug( 3, "AxKit FOP: checking dependency: $graphic" );
	$stats->{$graphic} = File::stat::stat( $graphic )
	  or warn "AxKit FOP: could not stat $graphic for " .
	    $xml_provider->key . ": $!" and next;
	if ( $stats->{$graphic}->mtime > $mtime ) {
	    AxKit::Debug( 4, "AxKit FOP: dependency newer: $graphic" );
	    return ( $cache, 1, $stats ); # cache not up-to-date
	}
    }

    ( $cache, 0, $stats ); # cache up-to-date
}

#-------------------------------------------------------------------------------
# Run FOP either from the command line or through JavaServer.
#-------------------------------------------------------------------------------
sub run_fop {
    my( $r, $style, $input, $config, $graphics, $stats, $xml_provider ) = @_;

    # Create temporary directory either in specified location or in tmpdir.
    my $javadir = $r->dir_config( 'AxFOPJavaServerDirectory' );
    my $dir = $javadir || File::Spec->tmpdir;
    my $tempdir = File::Temp::tempdir( 'axkit-fop-XXXXXX', DIR => $dir )
      or die "AxKit FOP: could not create tempdir under $dir: $!";
    AxKit::Debug( 8, "AxKit FOP: tempdir='$tempdir'" );
    $r->register_cleanup( sub { File::Path::rmtree( $tempdir ) } );
    if ( $javadir ) {
	my $stat = File::stat::stat( $tempdir )
	  or die "AxKit FOP: could not stat $tempdir: $!";
	my $mode = $stat->mode | S_IRWXG; # add group rwx permissions
	chmod( $mode, $tempdir )
	  or die "AxKit FOP: could not set permissions on $tempdir: $!";
    }

    # Write XSL-FO input for FOP to a file in tempdir.
    my $input_file = File::Spec->catfile( $tempdir, 'input.fo' );
    my $fh = Apache->gensym;
    open( $fh, ">$input_file" )
      or die "AxKit FOP: could not open $input_file: $!";
    print $fh $$input or die "AxKit FOP: could not write to $input_file: $!";
    close( $fh ) or die "AxKit FOP: could not close $input_file: $!";

    # If $config is a DOM object, then write it to a file in tempdir.
    my $config_file = $config;
    if ( ref $config ) {
	$config_file = File::Spec->catfile( $tempdir, 'config.xml' );
	AxKit::Debug( 8, "AxKit FOP: writing config file $config_file" );
	my $fh = Apache->gensym;
	open( $fh, ">$config_file" )
	  or die "AxKit FOP: could not open $config_file: $!";
	print $fh $config->toString
	  or die "AxKit FOP: could not write to $config_file: $!";
	close( $fh ) or die "AxKit FOP: could not close $config_file: $!";
    }

    my $output_file = File::Spec->catfile( $tempdir, "output.$style" );

    my $javadir     = $r->dir_config( 'AxFOPJavaServerDirectory' );
    my $port        = $r->dir_config( 'AxFOPJavaServerPort' );
    my $authfile    = $r->dir_config( 'AxFOPJavaServerAuthFile' );
    unless ( defined $javadir or defined $port or defined $authfile ) {
	# Call FOP via the command line.
	my $fop = $r->dir_config( 'AxFOPPath' ) || 'fop';
	my $command = "JAVA_OPTS=-Djava.awt.headless=true" .
	  " $fop -q $input_file -$style $output_file";
	$command .= " -c $config_file" if defined $config_file;
	AxKit::Debug( 1, "AxKit FOP: $command" );
	system( $command ) == 0 or die "AxKit FOP: fop died with status $?";
	return $output_file;
    }

    # Call FOP through JavaServer.
    require XML::ApacheFOP;
    my $debug = " port=$port" if defined $port;
    $debug .= " authfile=$authfile" if defined $authfile;
    AxKit::Debug( 8, "AxKit FOP: connecting to JavaServer$debug" );
    my $fop = XML::ApacheFOP->new( port => $port, authfile => $authfile );
    $fop->reset_image_cache
      if reset_image_cache( $r, $graphics, $stats, $xml_provider );
    $debug = defined $config_file ? ", c=$config_file" : ', c=undef';
    AxKit::Debug( 1, "AxKit FOP: running FOP via JavaServer: fo=$input_file, outfile=$output_file, rendertype=$style$debug" );
    $fop->fop( fo => $input_file, outfile => $output_file,
	       rendertype => $style, c => $config_file )
      or die "AxKit FOP: could run FOP via JavaServer: " . $fop->errstr;

    $output_file;
}

#-------------------------------------------------------------------------------
# Determine if FOP image cache needs to be reset.
#-------------------------------------------------------------------------------
sub reset_image_cache {
    my( $r, $graphics, $stats, $xml_provider ) = @_;

    # See if we should track images or not.
    my $image_cache = lc $r->dir_config( 'AxFOPImageCache' );
    if ( not defined $image_cache or $image_cache eq '' ) {
	$image_cache = 1000;
	AxKit::Debug( 8, "AxKit FOP: using default AxFOPImageCache $image_cache" );
    }
    elsif ( $image_cache eq 'off' ) {
	AxKit::Debug( 8, "AxKit FOP: AxFOPImageCache is Off;" .
		      " resetting FOP image cache" );
	return 1; # always reset image cache
    }
    elsif ( $image_cache eq 'on' ) {
	AxKit::Debug( 8, "AxKit FOP: AxFOPImageCache is On; " .
		      " leaving FOP image cache alone" );
	return 0; # never reset image cache
    }
    elsif ( $image_cache !~ m!^\d+$! ) {
	die "AxKit FOP: AxFOPImageCache invalid: $image_cache";
    }

    # See if we can track images or not.
    if ( $AxKit::Cfg->NoCache ) {
	AxKit::Debug( 8, "AxKit FOP: NoCache is On;" .
		      " resetting FOP image cache" );
	return 1; # always reset image cache
    }
    unless ( $AxKit::Cfg->DependencyChecks ) {
	AxKit::Debug( 8, "AxKit FOP: AxDependencyChecks Off;" .
		      " leaving FOP image cache alone" );
	return 0; # never reset image cache
    }

    # Lock and open DBM file for tracking images.
    my $dbm_file = $r->dir_config( 'AxFOPImageTrack' );
    $dbm_file = $AxKit::Cfg->CacheDir . '/AxFOPImageTrack'
      unless defined $dbm_file;
    AxKit::Debug( 10, "AxKit FOP: tracking images in $dbm_file" );
    my $fh = Apache->gensym;
    open( $fh, ">$dbm_file.lock" )
      or die "AxKit FOP: could not open $dbm_file.lock: $!";
    flock( $fh, LOCK_EX ) or die "AxKit FOP: could not lock $dbm_file.lock: $!";
    tie( my %cache, 'SDBM_File', $dbm_file, O_RDWR|O_CREAT, 0666 )
      or die "AxKit FOP: could not tie SDBM file $dbm_file: $!";

    # Find out if cache needs to be reset.
    my $reset = _reset_image_cache( \%cache, $graphics, $image_cache, $stats,
				    $xml_provider );
    %cache = () if $reset;

    untie %cache;
    close( $fh ) or die "AxKit FOP: could not close $dbm_file.lock: $!";
    $reset;
}

#-------------------------------------------------------------------------------
sub _reset_image_cache {
    my( $cache, $graphics, $image_cache, $stats, $xml_provider ) = @_;

    # Count number of images in cache and check if an image has changed.
    my $count = keys %$cache;
    my $reset = 0;
    foreach my $graphic ( @$graphics ) {
	my $stat;
	if ( $stats and exists $stats->{$graphic} ) {
	    $stat = $stats->{$graphic};
	}
	else {
	    $stat = File::stat::stat( $graphic )
	      or warn "AxKit FOP: could not stat $graphic for " .
		$xml_provider->key . ": $!";
	}

	# Note: length of SDBM key plus value may not exceed 1008 bytes.
	# A 64-bit timestamp is 20 bytes long (when converted to a string).
	my $key = substr( $graphic, -988 );
	++$count unless $cache->{$key};
	my $old_mtime = $cache->{$key};
	my $new_mtime = $stat ? $stat->mtime : undef;
	if ( not defined $old_mtime ) {
	    # image not yet in cache
	}
	elsif ( not defined $new_mtime ) {
	    AxKit::Debug( 1, "AxKit FOP: image disappeared: $graphic;" .
			  " resetting FOP image cache" );
	    ++$reset;
	}
	elsif ( $new_mtime != $old_mtime ) {
	    AxKit::Debug( 1, "AxKit FOP: image changed: $graphic;" .
			  " resetting FOP image cache" );
	    ++$reset;
	}
	$cache->{$key} = $new_mtime or delete $cache->{$key};
    }

    if ( not $reset and $count > $image_cache ) {
	AxKit::Debug( 1, "AxKit FOP: $count images over limit ($image_cache);" .
		      " resetting FOP image cache" );
	$reset = 1;
    }

    AxKit::Debug( 1, "AxKit FOP: not resetting FOP image cache" ) unless $reset;

    $reset;
}

1;

__END__

=head1 NAME

Apache::AxKit::Language::FOP - generate PDF using Apache FOP

=head1 SYNOPSIS

    AxAddStyleMap text/xslfo +Apache::AxKit::Language::FOP
    PerlSetVar AxFOPConfigFile /path/to/my/userconfig.xml
    PerlSetVar AxFOPPath       /usr/bin/fop

If you want to use JavaServer, specify one or more of:

    PerlSetVar AxFOPJavaServerPort      2000
    PerlSetVar AxFOPJavaServerAuthFile  /path/to/my/authfile
    PerlSetVar AxFOPJavaServerDirectory /my/special/directory

If you want to cache when PDF is generated dynamically with XSP:

    PerlSetVar AxFOPCache IfNoCache

When running under JavaServer, this module will reset FOP's image
cache when needed.  If you are using an older FOP that does not
support this feature, then specify the following to leave the image
cache alone:

    PerlSetVar AxFOPImageCache On

=head1 DESCRIPTION

This module uses Apache FOP to generate PDF by processing XSL-FO
files.  It can either call the command line tool "fop", or for better
performance, it can use the L<XML::ApacheFOP|XML::ApacheFOP> package
(which uses the L<Java.pm|Java> Perl package and JavaServer software).

For running FOP via the command line, you can specify the path to your
FOP executable with the directive AxFOPPath.  The default value is
simply "fop" which will get resolved by your shell using whatever PATH
is currently in effect.

To enable use of L<XML::ApacheFOP|XML::ApacheFOP>, specify one or more
AxFOPJavaServer directives as shown above.  Unspecified directives
will fall back to their default values defined in
L<XML::ApacheFOP|XML::ApacheFOP>.

If you need FOP to run with an external configuration file, specify
the location using the AxFOPConfigFile directive.  See
http://xmlgraphics.apache.org/fop/configuration.html for options you
can specify in an external configuration file.  For example, you can
specify a C<baseDir> to be used when resolving a relative URL from the
C<src> attribute of an C<fo:external-graphic>.

Multiple output render types are supported by FOP, to varying degrees:

    pdf: application/pdf
    mif: application/vnd.mif
    pcl: application/vnd.hp-PCL
    ps:  application/postscript
    txt: text/plain
    svg: image/svg+xml
    at:  text/xml

If the AxKit PreferredStyle matches one of the above styles, then that
style will be the render type output by this module.  If the
PreferredStyle does not match, then the output type will be pdf.

=head2 Caching

The standard AxKit caching mechanism automatically caches the output
in most cases (unless disabled with AxNoCache).  However, if your
content is being generated dynamically from an XSP script, then AxKit
will not cache the output.

To enable caching for XSP-generated content, set the configuration
directive AxFOPCache to "IfNoCache".  When this directive is set (and
caching is not disabled with AxNoCache), the output will be cached
separately by this module whenever AxKit has turned off its caching
(that is, when $AxKit::Cache->no_cache returns a true value).  The
cache file is managed using
L<Apache::AxKit::Cache|Apache::AxKit::Cache> with a key generated
using L<Digest::MD5|Digest::MD5> from the input content and the output
style.

To turn off separate caching by this module you can set S<"AxFOPCache
Off"> (this is the default).

The AxKit cache is not automatically cleaned.  If space is an issue,
such as when you are caching pages that are customized for individual
users, then you need to run a cron job or something similar to delete
stale files from the cache.  For example, see
L<Apache::AxKit::CacheCleaner|Apache::AxKit::CacheCleaner>.

=head2 Notes about C<fo:external-graphic>

You can use relative links to external graphics.  They will be
resolved relative to the directory of your source file.  If the
resulting file does not exist, then the relative link will be resolved
relative to the definition of C<baseDir> in the FOP configuration
file.  This behavior mimics that of FOP when called from the command
line--it resolves relative links first with respect to the current
directory and second with respect to C<baseDir>, which is set to the
location of your source file unless otherwise defined in the FOP
configuration file.

When using JavaServer be sure it has permission to locate and access
the external graphic file.

If you have not defined a C<baseDir> in a FOP configuration file, then
this module will create one with C<baseDir> set to the source
directory, in case FOP needs to resolve a relative link on its own.
(This feature may be redundant since the link to the external graphic
will have already been resolved by this module, but the feature
remains just in case FOP uses C<baseDir> for other reasons.)

Why does this module resolve the relative links instead of letting FOP
do it?  As explained below, FOP caches images between runs (when
called through JavaServer), and unfortunately it caches the image
based on the link I<before> it gets resolved, which can lead to the
wrong image being used if it has the same name but resides in a
different directory.  For example, if you have a document in one
directory that loads an image "logo.svg", and you have another
document in another directory that loads a different image "logo.svg",
then FOP will re-use the I<first> image.  An additional reason this
module resolves the links is because FOP might not run from the same
directory as your source file, especially if it is being run through
JavaServer, so FOP could load some random graphic relative to its
current directory instead of relative to your source file.

AxKit is informed of any external-graphics from your document so that
it can check the timestamp and recreate the output instead of using
its cache if a graphic file is newer than the cache file (unless you
have AxDependencyChecks Off).  This module will also perform these
dependency checks when you have turned on separate caching (unless you
have AxDependencyChecks Off).

When using XML::ApacheFOP with JavaServer, images are cached by FOP
between runs.  See
http://xmlgraphics.apache.org/fop/0.20.5/graphics.html#caching for
more information.  If you have not disabled caching with AxNoCache and
have not turned off AxDependencyChecks, then this module will keep
track of which external-graphics have been used and will reset FOP's
image cache whenever a graphic file changes or when FOP has cached
1000 images.  You can change the limit, for example:

  AxFOPImageCache 100

This module uses an SDBM file named F<AxFOPImageTrack> located in
AxKit's cache directory to keep track of the graphic files.  You can
specify a different location or file name, for example:

  AxFOPImageTrack /var/cache/axkit_fop_images

If you have disabled caching with AxNoCache, then this module will not
keep track of external-graphics, and it will reset FOP's image cache
before I<each> run.  You can force this behavior by specifying:

  AxFOPImageCache Off

The ability to reset the image cache is available starting with FOP
version 0.20.5.  If you have an earlier version, you can instruct this
module to leave the cache alone so as to avoid error messages in your
log:

  AxFOPImageCache On

This module will also leave the cache alone if you have turned off
AxDependencyChecks.  With an earlier version of FOP or when you
instruct this module to leave the cache alone, you should restart
JavaServer whenever you change an external-graphic file or whenever
you think it is running out of memory from caching too many images.
For example:

  /etc/init.d/javaserver restart
  invoke-rc.d javaserver restart # on Debian

=head1 CAVEATS

A temporary directory with a prefix of "axkit-fop-" is created on each
request (using L<File::Temp|File::Temp>) to contain the input and
output files for FOP.  This directory is readable/writable only by the
web server process and it is automatically deleted in the mod_perl
cleanup phase at the end of each request.

If you are using XML::ApacheFOP and JavaServer, then the JavaServer
process needs read/write permission on this temporary directory.  One
way to allow this permission is to run your JavaServer process under
the same user as your web server.

An alternative method, if your JavaServer is running as a different
user from your web server, is to run your JavaServer under a group in
which the web server belongs, and create a special temporary directory
owned by this group with the group sticky bit set so that
subdirectories will also be owned by the group.  Inform this module of
that directory with this directive:

    PerlSetVar AxFOPJavaServerDirectory /my/special/directory

This module will turn on the group read/write permissions on temporary
subdirectories created under this special directory.

Here is an example setup using Debian GNU/Linux commands (if someone
can generalize this example let me know).  Assume your JavaServer
process is running under the user "javaserv" and the group "javaserv",
and your web server is running under the user "www-data" and the group
"www-data".

Add your web server user to the JavaServer group:

    adduser www-data javaserv

Create a temporary directory with the group sticky bit set:

    install -m 2770 -o javaserv -g javaserv -d /var/cache/javaserver

Update your Apache configuration with this directive:

    PerlSetVar AxFOPJavaServerDirectory /var/cache/javaserver

=head1 SEE ALSO

L<AxKit|AxKit>,
L<XML::ApacheFOP|XML::ApacheFOP>,
L<Apache::AxKit::CacheCleaner|Apache::AxKit::CacheCleaner>

=head1 AUTHOR

Ken Neighbors <ken@nsds.com>

Based on the Apache::AxKit::Language::PassiveTex module, but severely
rewritten.

=head1 COPYRIGHT and LICENSE

Copyright (c) 2005 Ken Neighbors.  All rights reserved.  This module
is free software; you can redistribute it and/or modify it under the
same terms as Perl itself.

=cut
